package ru.dbotthepony.mc.otm.client.render

import com.mojang.blaze3d.vertex.PoseStack
import com.mojang.blaze3d.vertex.Tesselator
import com.mojang.blaze3d.vertex.VertexConsumer
import com.mojang.math.Matrix4f
import com.mojang.math.Vector3f
import net.minecraft.client.gui.Font
import net.minecraft.core.Vec3i
import net.minecraft.network.chat.Component
import net.minecraft.util.FormattedCharSequence
import ru.dbotthepony.mc.otm.core.*
import kotlin.math.roundToInt

val tesselator: Tesselator get() = Tesselator.getInstance()

fun VertexConsumer.normal(vector: Vector): VertexConsumer = normal(vector.x.toFloat(), vector.y.toFloat(), vector.z.toFloat())
fun VertexConsumer.vertex(matrix4f: Matrix4f, vector: Vector): VertexConsumer = vertex(matrix4f, vector.x.toFloat(), vector.y.toFloat(), vector.z.toFloat())
fun VertexConsumer.color(color: RGBAColor): VertexConsumer = color(color.red, color.green, color.blue, color.alpha)

fun PoseStack.translate(vector: Vector) = translate(vector.x, vector.y, vector.z)
fun PoseStack.translate(vector: Vec3i) = translate(vector.x.toDouble(), vector.y.toDouble(), vector.z.toDouble())
fun PoseStack.translate(vector: Vector3f) = last().pose().multiplyWithTranslation(vector.x(), vector.y(), vector.z())

fun PoseStack.rotateAroundPoint(point: Vector, axis: Vector, rotation: Float, isDegrees: Boolean = false) {
	val last = last()
	last.pose().rotateAroundPoint(point, axis, rotation, isDegrees)
	last.normal().mul(axis.rotateAroundThis(rotation, isDegrees))
}

fun PoseStack.rotateAroundPoint(point: Vector, rotation: IAngle) {
	val last = last()
	last.pose().rotateAroundPoint(point, rotation)
	// last.normal().mul(rotation.forward().rotateAroundThis(rotation))
}

fun PoseStack.rotateAroundPoint(point: Vector3f, axis: Vector, rotation: Float, isDegrees: Boolean = false) {
	val last = last()
	last.pose().rotateAroundPoint(point, axis, rotation, isDegrees)
	last.normal().mul(axis.rotateAroundThis(rotation, isDegrees))
}

fun PoseStack.rotateAroundPoint(point: Vector3f, rotation: IAngle) {
	val last = last()
	last.pose().rotateAroundPoint(point, rotation)
	// last.normal().mul(rotation.forward().rotateAroundThis(rotation))
}

fun PoseStack.translation(): Vector3f {
	return last().pose().getTranslation()
}

enum class TextAlign {
	TOP_LEFT,
	TOP_CENTER,
	TOP_RIGHT,

	CENTER_LEFT,
	CENTER_CENTER,
	CENTER_RIGHT,

	BOTTOM_LEFT,
	BOTTOM_CENTER,
	BOTTOM_RIGHT,
}

private fun Font.drawScaledDuckTyped(poseStack: PoseStack, text: Any, scale: Float, x: Float, y: Float, color: Int): Int {
	val translation = poseStack.translation()

	poseStack.pushPose()
	poseStack.translate(-translation)
	poseStack.scale(scale, scale, scale)
	val inv = 1f / scale
	poseStack.translate(-translation * inv)
	val size = drawDuckTyped(poseStack, text, x * inv, y * inv, color)
	poseStack.popPose()

	return size
}

private fun Font.drawDuckTyped(poseStack: PoseStack, text: Any, x: Float, y: Float, color: Int): Int {
	return when (text) {
		is Component -> draw(poseStack, text, x, y, color)
		is String -> draw(poseStack, text, x, y, color)
		is FormattedCharSequence -> draw(poseStack, text, x, y, color)
		else -> throw ClassCastException(text::class.qualifiedName)
	}
}

private fun Font.widthDuckTyped(text: Any): Int {
	return when (text) {
		is Component -> width(text)
		is String -> width(text)
		is FormattedCharSequence -> width(text)
		else -> throw ClassCastException(text::class.qualifiedName)
	}
}

private fun Font.drawAlignedDuckTyped(poseStack: PoseStack, text: Any, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return when (align) {
		TextAlign.TOP_LEFT -> drawDuckTyped(poseStack, text, x, y, color)
		TextAlign.TOP_CENTER -> drawDuckTyped(poseStack, text, (x - widthDuckTyped(text) / 2f).roundToInt().toFloat(), y, color)
		TextAlign.TOP_RIGHT -> drawDuckTyped(poseStack, text, (x - widthDuckTyped(text)).roundToInt().toFloat(), y, color)

		TextAlign.CENTER_LEFT -> drawDuckTyped(poseStack, text, x, (y - lineHeight / 2f).roundToInt().toFloat(), color)
		TextAlign.CENTER_CENTER -> drawDuckTyped(poseStack, text, (x - widthDuckTyped(text) / 2f).roundToInt().toFloat(), (y - lineHeight / 2f).roundToInt().toFloat(), color)
		TextAlign.CENTER_RIGHT -> drawDuckTyped(poseStack, text, (x - widthDuckTyped(text)).roundToInt().toFloat(), (y - lineHeight / 2f).roundToInt().toFloat(), color)

		TextAlign.BOTTOM_LEFT -> drawDuckTyped(poseStack, text, x, (y - lineHeight).roundToInt().toFloat(), color)
		TextAlign.BOTTOM_CENTER -> drawDuckTyped(poseStack, text, (x - widthDuckTyped(text) / 2f).roundToInt().toFloat(), (y - lineHeight).roundToInt().toFloat(), color)
		TextAlign.BOTTOM_RIGHT -> drawDuckTyped(poseStack, text, (x - widthDuckTyped(text)).roundToInt().toFloat(), (y - lineHeight).roundToInt().toFloat(), color)
	}
}

private fun Font.drawScaledAlignedDuckTyped(poseStack: PoseStack, text: Any, scale: Float, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return when (align) {
		TextAlign.TOP_LEFT -> drawScaledDuckTyped(poseStack, text, scale, x, y, color)
		TextAlign.TOP_CENTER -> drawScaledDuckTyped(poseStack, text, scale, (x - widthDuckTyped(text) * scale / 2f).roundToInt().toFloat(), y, color)
		TextAlign.TOP_RIGHT -> drawScaledDuckTyped(poseStack, text, scale, (x - widthDuckTyped(text) * scale).roundToInt().toFloat(), y, color)

		TextAlign.CENTER_LEFT -> drawScaledDuckTyped(poseStack, text, scale, x, (y - lineHeight / 2f * scale).roundToInt().toFloat(), color)
		TextAlign.CENTER_CENTER -> drawScaledDuckTyped(poseStack, text, scale, (x - widthDuckTyped(text) * scale / 2f).roundToInt().toFloat(), (y - lineHeight * scale / 2f).roundToInt().toFloat(), color)
		TextAlign.CENTER_RIGHT -> drawScaledDuckTyped(poseStack, text, scale, (x - widthDuckTyped(text) * scale).roundToInt().toFloat(), (y - lineHeight * scale / 2f).roundToInt().toFloat(), color)

		TextAlign.BOTTOM_LEFT -> drawScaledDuckTyped(poseStack, text, scale, x, (y - lineHeight * scale).roundToInt().toFloat(), color)
		TextAlign.BOTTOM_CENTER -> drawScaledDuckTyped(poseStack, text, scale, (x - widthDuckTyped(text) * scale / 2f).roundToInt().toFloat(), (y - lineHeight * scale).roundToInt().toFloat(), color)
		TextAlign.BOTTOM_RIGHT -> drawScaledDuckTyped(poseStack, text, scale, (x - widthDuckTyped(text) * scale).roundToInt().toFloat(), (y - lineHeight * scale).roundToInt().toFloat(), color)
	}
}

fun Font.drawAligned(poseStack: PoseStack, text: String, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return drawAlignedDuckTyped(poseStack, text, align, x, y, color)
}

fun Font.drawAligned(poseStack: PoseStack, text: Component, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return drawAlignedDuckTyped(poseStack, text, align, x, y, color)
}

fun Font.drawAligned(poseStack: PoseStack, text: FormattedCharSequence, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return drawAlignedDuckTyped(poseStack, text, align, x, y, color)
}

fun Font.drawAligned(poseStack: PoseStack, text: String, align: TextAlign, x: Float, y: Float, color: RGBAColor): Int {
	return drawAligned(poseStack, text, align, x, y, color.toInt())
}

fun Font.drawAligned(poseStack: PoseStack, text: Component, align: TextAlign, x: Float, y: Float, color: RGBAColor): Int {
	return drawAligned(poseStack, text, align, x, y, color.toInt())
}

fun Font.drawAligned(poseStack: PoseStack, text: FormattedCharSequence, align: TextAlign, x: Float, y: Float, color: RGBAColor): Int {
	return drawAligned(poseStack, text, align, x, y, color.toInt())
}

fun Font.drawScaledAligned(poseStack: PoseStack, text: String, scale: Float, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return drawScaledAlignedDuckTyped(poseStack, text, scale, align, x, y, color)
}

fun Font.drawScaledAligned(poseStack: PoseStack, text: Component, scale: Float, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return drawScaledAlignedDuckTyped(poseStack, text, scale, align, x, y, color)
}

fun Font.drawScaledAligned(poseStack: PoseStack, text: FormattedCharSequence, scale: Float, align: TextAlign, x: Float, y: Float, color: Int): Int {
	return drawScaledAlignedDuckTyped(poseStack, text, scale, align, x, y, color)
}

fun Font.drawScaledAligned(poseStack: PoseStack, text: String, scale: Float, align: TextAlign, x: Float, y: Float, color: RGBAColor): Int {
	return drawScaledAligned(poseStack, text, scale, align, x, y, color.toInt())
}

fun Font.drawScaledAligned(poseStack: PoseStack, text: Component, scale: Float, align: TextAlign, x: Float, y: Float, color: RGBAColor): Int {
	return drawScaledAligned(poseStack, text, scale, align, x, y, color.toInt())
}

fun Font.drawScaledAligned(poseStack: PoseStack, text: FormattedCharSequence, scale: Float, align: TextAlign, x: Float, y: Float, color: RGBAColor): Int {
	return drawScaledAligned(poseStack, text, scale, align, x, y, color.toInt())
}
