package ru.dbotthepony.mc.otm.item

import net.minecraft.tags.BlockTags
import net.minecraft.world.InteractionResult
import net.minecraft.world.entity.vehicle.AbstractMinecart
import net.minecraft.world.item.DyeColor
import net.minecraft.world.item.Item
import net.minecraft.world.item.context.UseOnContext
import net.minecraft.world.level.block.BaseRailBlock
import net.minecraft.world.level.block.state.properties.RailShape
import net.minecraft.world.level.gameevent.GameEvent
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.entity.MinecartCargoCrate
import ru.dbotthepony.mc.otm.registry.MEntityTypes

class MinecartCargoCrateItem(val color: DyeColor?) : Item(Properties().stacksTo(64).tab(OverdriveThatMatters.INSTANCE.CREATIVE_TAB)) {
	override fun useOn(useOnContext: UseOnContext): InteractionResult {
		val level = useOnContext.level
		val blockpos = useOnContext.clickedPos
		val blockstate = level.getBlockState(blockpos)

		if (!blockstate.`is`(BlockTags.RAILS)) {
			return InteractionResult.FAIL
		}

		val itemstack = useOnContext.itemInHand

		if (!level.isClientSide) {
			val railshape =
				if (blockstate.block is BaseRailBlock)
					(blockstate.block as BaseRailBlock).getRailDirection(blockstate, level, blockpos, null)
				else
					RailShape.NORTH_SOUTH

			val d0 = if (railshape.isAscending) 0.5 else 0.0
			val type = MEntityTypes.CARGO_CRATE_MINECARTS[color] ?: throw NullPointerException("Unable to find minecart entity type for color $color!")
			val entity = MinecartCargoCrate(type, color, level, blockpos.x.toDouble() + 0.5, blockpos.y.toDouble() + 0.0625 + d0, blockpos.z.toDouble() + 0.5)

			if (itemstack.hasCustomHoverName()) {
				entity.customName = itemstack.hoverName
			}

			level.addFreshEntity(entity)
			level.gameEvent(GameEvent.ENTITY_PLACE, blockpos, GameEvent.Context.of(useOnContext.player, level.getBlockState(blockpos.below())))
		}

		if (useOnContext.player?.abilities?.instabuild != true) {
			itemstack.shrink(1)
		}

		return InteractionResult.sidedSuccess(level.isClientSide)
	}
}
