package ru.dbotthepony.mc.otm.network

import it.unimi.dsi.fastutil.ints.Int2ObjectAVLTreeMap
import it.unimi.dsi.fastutil.ints.Int2ObjectArrayMap
import it.unimi.dsi.fastutil.ints.IntAVLTreeSet
import it.unimi.dsi.fastutil.ints.IntArrayList
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import it.unimi.dsi.fastutil.objects.Object2ObjectArrayMap
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.world.entity.player.Player
import net.minecraft.world.inventory.AbstractContainerMenu
import net.minecraft.world.inventory.Slot
import net.minecraft.world.item.ItemStack
import net.minecraftforge.network.NetworkDirection
import net.minecraftforge.network.NetworkEvent
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.block.entity.storage.ItemMonitorPlayerSettings
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.container.ItemFilterSlotPacket
import ru.dbotthepony.mc.otm.core.TextComponent
import ru.dbotthepony.mc.otm.menu.*
import ru.dbotthepony.mc.otm.menu.data.*
import ru.dbotthepony.mc.otm.menu.widget.BooleanPlayerInputPacket
import ru.dbotthepony.mc.otm.menu.widget.NumberPlayerInputPacket
import ru.dbotthepony.mc.otm.menu.widget.OneWayPlayerInputPacket
import java.io.ByteArrayInputStream
import java.util.function.Supplier

class MenuFieldPacket(val bytes: ByteArray, val length: Int) : MatteryPacket {
	constructor(stream: FastByteArrayOutputStream) : this(stream.array, stream.length)

	override fun write(buff: FriendlyByteBuf) {
		buff.writeBytes(bytes, 0, length)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.get().enqueueWork {
			(minecraft.player?.containerMenu as? MatteryMenu)?.mSynchronizer?.applyNetworkPayload(ByteArrayInputStream(bytes, 0, length))
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): MenuFieldPacket {
			val readable = buff.readableBytes()
			return MenuFieldPacket(ByteArray(readable).also(buff::readBytes), readable)
		}
	}
}

class SetCarriedPacket(val item: ItemStack) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeItem(item)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork { minecraft.player?.containerMenu?.carried = item }
	}

	companion object {
		fun read(buff: FriendlyByteBuf): SetCarriedPacket {
			return SetCarriedPacket(buff.readItem())
		}
	}
}

object MenuNetworkChannel : MatteryNetworkChannel(
	version = "1",
	name = "menu"
) {
	fun register() {
		add(SetCarriedPacket::class.java, SetCarriedPacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)
		add(ItemFilterSlotPacket::class.java, ItemFilterSlotPacket.Companion::read)

		// networked view
		add(ClearItemViewPacket::class.java, ClearItemViewPacket::read, NetworkDirection.PLAY_TO_CLIENT)
		add(ItemViewInteractPacket::class.java, ItemViewInteractPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)
		add(StackAddPacket::class.java, StackAddPacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)
		add(StackChangePacket::class.java, StackChangePacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)
		add(StackRemovePacket::class.java, StackRemovePacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)

		// container data
		add(MenuFieldPacket::class.java, MenuFieldPacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)

		// Player input in menus
		add(NumberPlayerInputPacket::class.java, NumberPlayerInputPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)
		add(OneWayPlayerInputPacket::class.java, OneWayPlayerInputPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)
		add(BooleanPlayerInputPacket::class.java, BooleanPlayerInputPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)

		// menu specific

		// Item monitor
		add(ItemMonitorPlayerSettings::class.java, ItemMonitorPlayerSettings.Companion::read)

		// matter panel menu
		add(CancelTaskPacket::class.java, CancelTaskPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)
		add(PatternsChangePacket::class.java, PatternsChangePacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)
		add(TasksChangePacket::class.java, TasksChangePacket.Companion::read, NetworkDirection.PLAY_TO_CLIENT)
		add(ReplicationRequestPacket::class.java, ReplicationRequestPacket.Companion::read, NetworkDirection.PLAY_TO_SERVER)
	}
}
